#define EXTENDED_VERSION "3.2.0-git (git e1855213f333e96f65b952676cc03f93942b1e57)"
#define PREMIUM_EXTENDED_VERSION "no"
#define PACKAGE_VERSION_TYPE "stable"
#define SOURCE_OF_INSTALLATION "git e1855213f333e96f65b952676cc03f93942b1e57"
